#lang racket

;; space-invaders-3.rkt

; in this version, we will add for-test: methods to facilitate testing
; first look at the probe for bomb
; then, the probe for world

(require rackunit)
(require 2htdp/universe)
(require 2htdp/image)
(require "extras.rkt")


;; Press space to drop a new bomb.  
;; Bombs fall at a constant rate. 
;; Bombs are draggable. 

;; Helicopter just rises at a constant rate.

;; start with (run framerate).  Typically: (run 0.25)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; CONSTANTS

(define CANVAS-WIDTH 200)
(define CANVAS-HEIGHT 400)

(define EMPTY-CANVAS (empty-scene 200 400))

;; arbitrary choice
(define NEW-BOMB-X-POS 75)  
(define NEW-BOMB-Y-POS 0)
(define NEW-BOMB-RADIUS 10)

(define HELI-INITIAL-X 100)
(define HELI-INITIAL-Y 300)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; INTERFACE

(define WorldObj<%>
  (interface ()

    ; -> WorldObj<%>
    ; GIVEN: no arguments
    ; RETURNS: the state of this object that should follow a
    ; tick. 
    after-tick          

    ; Number Number MouseEvent -> WorldObj<%>
    ; GIVEN: a location and a mouse event
    ; RETURNS: the state of this object that should follow the
    ; mouse event.
    after-mouse-event   

    ; KeyEvent -> WorldObj<%>
    ; GIVEN: a key event
    ; RETURNS: the state of this object that should follow the
    ; key event.
    after-key-event     

    ; Scene -> Scene
    ; GIVEN: a scene
    ; RETURNS: a scene like the given one, but with this object
    ; painted on it.
    to-scene
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Here's what a class definition looks like:

;; classes are like data definitions.  They should have a purpose statement
;; describing what information they are supposed to represent, and
;; interpretations of the fields describing the meaning of each piece of data.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; A World is a (new World% [heli Helicopter] [bombs ListOf<Bomb>])
;; interpretation: represents a world, containing a helicopter and some bombs.
(define World%               ; the % is silent
  (class* object% (WorldObj<%>)         
    ;; what you need to create an object of this class
    (init-field heli)        ; a Helicopter   --the helicopter in the game
    (init-field bombs)       ; a ListOf<Bomb> -- the list of bombs that the UFO has dropped.             

    ;; you can declare private fields, too; see below
    
    ;; this line is mandatory magic   
    (super-new)

    ;; after-tick : -> World
    ;; RETURNS: A world like this one, but as it should be after a tick
    ;; STRATEGY: function composition
    (define/public (after-tick)
      (new World%
        [heli    (send heli after-tick)] 
        [bombs   (map
                   (lambda (bomb) (send bomb after-tick))
                   bombs)]))
    
    ;; after-mouse-event : Number Number MouseEvent -> World
    ;; RETURNS: A world like this one, but as it should be after the
    ;; given mouse event.
    ;; STRATEGY: function composition
    (define/public (after-mouse-event x y evt)
      (new World%
           [heli (send heli after-mouse-event x y evt)]
           [bombs (map
                   (lambda (bomb) (send bomb after-mouse-event x y evt))
                   bombs)]))
    
    ;; after-key-event : KeyEvent -> World
    ;; RETURNS: A world like this one, but as it should be after the
    ;; given key event.
    ;; STRATEGY: structural decomposition on KeyEvent (kev)
    ;; DETAILS: on space, drop a bomb, otherwise ignore
    ;; NOTE: here I was too lazy to provide a proper data definition
    ;; for SpaceInvadersKeyEvent.  -2 for me.  You should do better.
    (define/public (after-key-event kev)
      (cond
        [(key=? kev " ")
         (new World%
              [heli heli]
              [bombs (cons (make-bomb) bombs)])]
        [else this]))    
    
    ;; to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this world painted
    ;; on it.
    ;; STRATEGY: function composition
    (define/public (to-scene scene0)
      (local
        ;; first add the helicopter to the scene
        ((define scene-with-helicopter (send heli to-scene scene0)))
        ;; then tell each bomb to add itself to the scene
        (foldr
         (lambda (bomb scene)
          (send bomb to-scene scene))
         scene-with-helicopter
         bombs)))

    ;; -> Heli
    (define/public (get-heli) heli)
    ;; -> ListOf<Bomb>
    (define/public (get-bombs) bombs)
 
    ))          ; this is one of the few times we will allow you
                ; to have right parens on a line by themselves

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; A Bomb is a (new Bomb% [x Number][y Number]
;;                        [radius Number][selected? Boolean])
;; A Bomb represents a bomb.
(define Bomb%
  (class* object% (WorldObj<%>) 
    (init-field
     x            ; the bomb's x position, in pixels, 
                  ; relative to the upper-left corner of the canvas
     y )          ; the bomb's y position
    (init-field r)   ; the bomb's radius
    (init-field [selected? false])   ; the bomb's selected? status-
                                     ; initially false.
     
    
    ;; private data for objects of this class.
    ;; these can depend on the init-fields.
    (field [IMG (circle r "solid" "red")])     ; image for displaying the bomb
    (field [BOMBSPEED 4])                      ; the bomb's speed, in pixels/tick
       
    (super-new)
    
    ;; after-tick : -> Bomb
    ;; RETURNS: A bomb like this one, but as it should be after a tick
    (define/public (after-tick)
      (new Bomb% [x x][y (+ y BOMBSPEED)][r r][selected? selected?]))
    
    ;; after-key-event : KeyEvent -> Bomb
    ;; RETURNS: A world like this one, but as it should be after the
    ;; given key event.
    ;; DETAILS: a bomb ignores key events
    ;; STRATEGY: domain knowledge
    (define/public (after-key-event kev)
      this)      
    
    ; after-mouse-event : Number Number MouseEvent -> Bomb
    ; GIVEN: the location of a mouse event, and the mouse event
    ; RETURNS: the bomb that should follow this one after the given
    ; mouse event.
    ; STRATEGY: cases on evt : MouseEvent
    (define/public (after-mouse-event mouse-x mouse-y evt)
      (cond
        [(mouse=? evt "button-down")
         (send this bomb-after-button-down mouse-x mouse-y)]
        [(mouse=? evt "drag") 
         (send this bomb-after-drag mouse-x mouse-y)]
        [(mouse=? evt "button-up")
         (send this bomb-after-button-up)]
        [else this]))

    
    ;; bomb-after-button-down : Number Number -> Bomb
    ;; GIVEN: the location of a mouse event
    ;; RETURNS: the bomb that should follow this one after a button
    ;; down at the given location
    ;; DETAILS:  If the event is inside
    ;; the bomb, returns a bomb just like this bomb, except that it is
    ;; selected.  Otherwise returns the bomb unchanged.
    (define/public (bomb-after-button-down mouse-x mouse-y)
      (if (send this in-bomb? mouse-x mouse-y)
        (new Bomb% [x x][y y][r r][selected? true])
        this))

    ;; bomb-after-drag : Number Number -> Bomb
    ;; GIVEN: the location of a mouse event
    ;; RETURNS: the bomb that should follow this one after a drag at
    ;; the given location 
    ;; DETAILS: if bomb is selected, move the bomb to the mouse location,
    ;; otherwise ignore.
    (define/public (bomb-after-drag mouse-x mouse-y)
      (if selected?
        (new Bomb% [x mouse-x][y mouse-y][r r][selected? true])
        this))


    ;; bomb-after-button-up : -> Bomb
    ;; RETURNS: the bomb that should follow this one after a button-up
    ;; DETAILS: button-up unselects all bombs
    ;; STRATEGY: domain knowledge  (NOTE: new doesn't count)
    (define/public (bomb-after-button-up)
      (new Bomb% [x x][y y][r r][selected? false]))

    ;; to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this bomb painted
    ;; on it.
    (define/public (to-scene scene)
      (place-image IMG x y scene))
    
    ;; in-bomb? : Number Number -> Boolean
    ;; GIVEN: a location on the canvas
    ;; RETURNS: true iff the location is inside this bomb.
    (define/public (in-bomb? other-x other-y)
      (<= (+ (sqr (- x other-x)) (sqr (- y other-y)))
          (sqr r)))

    ;; test methods, to probe the bomb state.  Note that we don't have
    ;; a probe for radius.
    ;; -> Int
    (define/public (get-x) x)
    ;; -> Int
    (define/public (get-y) y)
    ;; -> Boolean
    (define/public (get-selected?) selected?)

    ;; -> (list Int Int Boolean)
    (define/public (get-state) (list x y selected?))

    ))

;; GP: add "e" command to expand all the bombs.  Add probe and test cases.

;; A Heli is a (new Heli% [x Number][y Number])
;; A Heli represents a helicopter
;; in this version, the helicopter just rises slowly.
;; in a real system, the helicopter would behave quite differently from a bomb.
(define Heli%
  (class* object% (WorldObj<%>)
    (init-field
     x            ; the heli's x position
     y            ; the heli's y position
     )
    
    ;; private data for objects of this class.
    ;; these can depend on the init-fields.
    (field [HELI-IMG                             ; image for displaying the heli
             (square 20 "outline" "green")])
    (field [HELI-SPEED -2])                    ; the heli's speed, in pixels/tick
   
    (super-new)
    
    ;; after-tick : -> Heli
    ;; RETURNS: A heli like this one, but as it should be after a tick
    ;; DETAILS: the helicopter moves vertically by HELI-SPEED
    (define/public (after-tick)
      (new Heli% [x x][y (+ y HELI-SPEED)]))
    
    ; after-mouse-event : Number Number MouseEvent -> Heli
    ; GIVEN: the location of a mouse event, and the mouse event
    ; RETURNS: the bomb that should follow this one after the given
    ; mouse event.
    ; DETAILS: the helicopter ignores mouse events
    (define/public (after-mouse-event x y evt)
     this)

    ;; after-key-event : KeyEvent -> Heli
    ;; RETURNS: A world like this one, but as it should be after the
    ;; given key event.
    ;; DETAILS: a heli ignores key events
    (define/public (after-key-event kev)
      this)    
    
    ;; to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this heli painted
    ;; on it.
    (define/public (to-scene scene)
      (place-image HELI-IMG x y scene))   

    ;; test methods, to probe the heli state.
    (define/public (get-x) x)
    (define/public (get-y) y)

    ))

;; make-bomb: -> Bomb
;; RETURNS: a new bomb at the center top of the screen
(define (make-bomb)
  (new Bomb% [x NEW-BOMB-X-POS][y NEW-BOMB-Y-POS][r NEW-BOMB-RADIUS]))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; setting up the world:

;; initial-world : -> World
;; RETURNS: a world with a helicopter and no bombs
(define (initial-world)
  (new World% 
       [heli (new Heli% [x HELI-INITIAL-X][y HELI-INITIAL-Y])]
       [bombs empty]))

; run : Number -> World
; GIVEN: a frame rate, in secs/tick
; EFFECT: runs an initial world at the given frame rate
; RETURNS: the final state of the world
(define (run rate)
  (big-bang (initial-world)
            (on-tick
              (lambda (w) (send w after-tick))
              rate)
            (on-draw
              (lambda (w) (send w to-scene EMPTY-CANVAS)))
            (on-key
              (lambda (w kev) (send w after-key-event kev)))
            (on-mouse
              (lambda (w x y evt) (send w after-mouse-event x y
                                    evt)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Testing Framework

;; bomb-equal? : Bomb Bomb -> Boolean
;; GIVEN: two bombs
;; RETURNS: true iff they have the same x, y, and selected? fields
;; STRATEGY: morally, this is SD on the two bombs
(define (bomb-equal? b1 b2)
  (and
    (= 
      (send b1 get-x)
      (send b2 get-x))
    (=
      (send b1 get-y)
      (send b2 get-y))   
    (equal?      
      (send b1 get-selected?)
      (send b2 get-selected?))))

;; Heli Heli -> Boolean
(define (heli-equal? heli1 heli2)
  (and
    (= (send heli1 get-x) (send heli2 get-x))
    (= (send heli1 get-y) (send heli2 get-y))))

(define (world-equal? w1 w2)
  (and
     (heli-equal? (send w1 get-heli) (send w2 get-heli))
     (andmap
      (lambda (b1 b2) (bomb-equal? b1 b2))
      (send w1 get-bombs)
      (send w2 get-bombs))))

;; here we've used the 2-argument version of andmap, which is available in
;; #lang racket.

;; In what way does world-equal? equate two worlds that might be different?

;; In what way does world-equal? fail to equate two worlds that might
;; reasonably considered similar? 

(begin-for-test
  
  ;; using check-equal? as we would in the functional version
  (local
    ((define b1 (new Bomb% [x 20][y 30][r 5])))
    (check-equal? 
     (send b1 after-tick)
     (new Bomb% 
          [x 20][y (+ 30 4)]
          [r 5][selected? false])
     "Surprise!"))
                  
 

  (local
    ((define b1 (new Bomb% [x 20][y 30][r 5]))
     (define b2 (send b1 after-mouse-event 21 31 "button-down"))
     (define b3 (send b1 after-tick)))
    ;; bomb-equal? doesn't look at radius.  selected? defaults to false
    (check bomb-equal? b1 (new Bomb% [x 20][y 30][r 1000]))
    (check bomb-equal? b2 (new Bomb% 
                               [x 20][y 30]
                               [r 10][selected? true]))
    (check bomb-equal? b3 (new Bomb% 
                               [x 20][y (+ 30 4)]
                               [r 1000][selected? false])))

  )






