#lang racket

;; different versions of unit tests

(require rackunit)
(require 2htdp/universe)
(require 2htdp/image)
(require rackunit/text-ui)

(define NEW-BOMB-X-POS 75)  
(define NEW-BOMB-Y-POS 0)
(define NEW-BOMB-RADIUS 10)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Here's what a class definition looks like:

;; classes are like data definitions.  They should have a purpose statement
;; describing what information they are supposed to represent, and
;; interpretations of the fields describing the meaning of each piece of data.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; A Bomb is a (new Bomb% [x Number][y Number]
;;                        [radius Number][selected? Boolean])
;; A Bomb represents a bomb.
(define Bomb%
  (class* object% () 
    (init-field x y)
    ;; the position of the center of the bomb, in pixels, relative to
    ;; the upper-left corner of the canvas
    (init-field r)   ; the bomb's radius, in pixels
    (init-field [selected? false])   ; the bomb's selected? status-
                                     ; initially false.
     
    
    ;; private data for objects of this class.
    ;; these can depend on the init-fields.
    (field [IMG (circle r "solid" "red")])     ; image for displaying the bomb
    (field [BOMBSPEED 4])                      ; the bomb's speed, in pixels/tick
       
    (super-new)
    
    ;; after-tick : -> Bomb
    ;; RETURNS: A bomb like this one, but as it should be after a tick
    (define/public (after-tick)
      (new Bomb% [x x][y (+ y BOMBSPEED)][r r][selected? selected?]))
    
    ;; after-key-event : KeyEvent -> Bomb
    ;; RETURNS: A world like this one, but as it should be after the
    ;; given key event.
    ;; DETAILS: a bomb ignores key events
    ;; STRATEGY: domain knowledge
    (define/public (after-key-event kev)
      this)      
    
    ; after-mouse-event : Number Number MouseEvent -> Bomb
    ; GIVEN: the location of a mouse event, and the mouse event
    ; RETURNS: the bomb that should follow this one after the given
    ; mouse event.
    ; STRATEGY: cases on evt : MouseEvent
    (define/public (after-mouse-event mouse-x mouse-y evt)
      (cond
        [(mouse=? evt "button-down")
         (send this bomb-after-button-down mouse-x mouse-y)]
        [(mouse=? evt "drag") 
         (send this bomb-after-drag mouse-x mouse-y)]
        [(mouse=? evt "button-up")
         (send this bomb-after-button-up)]
        [else this]))

    
    ;; bomb-after-button-down : Number Number -> Bomb
    ;; GIVEN: the location of a mouse event
    ;; RETURNS: the bomb that should follow this one after a button
    ;; down at the given location
    ;; DETAILS:  If the event is inside
    ;; the bomb, returns a bomb just like this bomb, except that it is
    ;; selected.  Otherwise returns the bomb unchanged.
    (define/public (bomb-after-button-down mouse-x mouse-y)
      (if (send this in-bomb? mouse-x mouse-y)
        (new Bomb% [x x][y y][r r][selected? true])
        this))

    ;; bomb-after-drag : Number Number -> Bomb
    ;; GIVEN: the location of a mouse event
    ;; RETURNS: the bomb that should follow this one after a drag at
    ;; the given location 
    ;; DETAILS: if bomb is selected, move the bomb to the mouse location,
    ;; otherwise ignore.
    (define/public (bomb-after-drag mouse-x mouse-y)
      (if selected?
        (new Bomb% [x mouse-x][y mouse-y][r r][selected? true])
        this))

    ;; bomb-after-button-up : -> Bomb
    ;; RETURNS: the bomb that should follow this one after a button-up
    ;; DETAILS: button-up unselects all bombs
    ;; STRATEGY: domain knowledge  (NOTE: new doesn't count)
    (define/public (bomb-after-button-up)
      (new Bomb% [x x][y y][r r][selected? false]))
    
    ;; in-bomb? : Number Number -> Boolean
    ;; GIVEN: a location on the canvas
    ;; RETURNS: true iff the location is inside this bomb.
    (define/public (in-bomb? other-x other-y)
      (<= (+ (sqr (- x other-x)) (sqr (- y other-y)))
          (sqr r)))

    ;; here are several versions of probe methods:

    ;;;;;;;;;;;;;;;;

    ;; probe for individual field values.  This is the most primitive
    ;; kind of probe. By using distinctive names, we dissuade the
    ;; student from using these methods in ordinary code.

    ;; -> Int
    (define/public (for-test:bomb-x) x)
    ;; -> Int
    (define/public (for-test:bomb-y) y)
    ;; -> Boolean
    (define/public (for-test:bomb-selected?) selected?)

    ;;;;;;;;;;;;;;;;

    ;; single probe returning a list of the field values.  This is
    ;; easier, but it's confusing to the student because it returns a
    ;; heterogeneous list.  By rights, we should create a struct to
    ;; hold these values, but that seems like overkill.

    ;; -> (list Int Int Boolean)
    (define/public (for-test:bomb-state) (list x y selected?))

    ;;;;;;;;;;;;;;;;

    ;; another possibility is to have a has-properties METHOD

    ;; Int Int Boolean -> Boolean
    (define/public (for-test:bomb-has-properties xx yy bb)
      (and
        (= x xx)
        (= y yy)
        (equal? selected? bb)))

    ;;;;;;;;;;;;;;;;

    ;; another possibility is to run the unit tests INTERNALLY
    ;; this has the advantage that BOMBSPEED is in scope

    ;; -> Any
    (define/public (for-test:run-unit-tests)
      (local
        ((define b1 (new Bomb% [x 20][y 30][r 5]))
         (define b2 (send b1 after-mouse-event 21 31 "button-down"))
         (define b3 (send b1 after-tick)))
        (check-true (bomb-has-properties b1 20 30 false))
        (check-true (bomb-has-properties b2 20 30 true))
        (check-true (bomb-has-properties b3 20 (+ 30 BOMBSPEED) false))))


    ))

;; make-bomb: -> Bomb
;; RETURNS: a new bomb at the center top of the screen
(define (make-bomb)
  (new Bomb% [x NEW-BOMB-X-POS][y NEW-BOMB-Y-POS][r NEW-BOMB-RADIUS]))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; bomb-has-properties : Bomb Int Int Boolean -> Boolean
;; GIVEN: a bomb, and proposed values for its x, y, and selected?
;; fields
;; RETURNS: true iff the bomb's fields have those values
(define (bomb-has-properties b x y selected?)
  (and
    (= (send b for-test:bomb-x) x)
    (= (send b for-test:bomb-y) y)
    (equal? (send b for-test:bomb-selected?) selected?)))

;; bomb-properties :  Bomb -> (list Int Int Boolean)
(define (bomb-properties b)
  (list 
    (send b for-test:bomb-x)
    (send b for-test:bomb-y)
    (send b for-test:bomb-selected?)))

(define-test-suite bomb-tests

  ;; version with bomb-has-properties function
  (local
    ((define b1 (new Bomb% [x 20][y 30][r 5]))
     (define b2 (send b1 after-mouse-event 21 31 "button-down"))
     (define b3 (send b1 after-tick)))
    (check-true (bomb-has-properties b1 20 30 false))
    (check-true (bomb-has-properties b2 20 30 true))
    (check-true (bomb-has-properties b3 20 (+ 30 4) false)))

  ;; version with has-properties method
  (local
    ((define b1 (new Bomb% [x 20][y 30][r 5]))
     (define b2 (send b1 after-mouse-event 21 31 "button-down"))
     (define b3 (send b1 after-tick)))
    (check-true (send b1 for-test:bomb-has-properties 20 30 false))
    (check-true (send b2 for-test:bomb-has-properties 20 30 true))
    (check-true (send b3 for-test:bomb-has-properties 20 (+ 30 4) false)))

  ;; version using bomb-properties function
  (local
    ((define b1 (new Bomb% [x 20][y 30][r 5]))
     (define b2 (send b1 after-mouse-event 21 31 "button-down"))
     (define b3 (send b1 after-tick)))
    (check-equal? (bomb-properties b1) (list 20 30 false))
    (check-equal? (bomb-properties b2) (list 20 30 true))
    (check-equal? (bomb-properties b3) (list 20 (+ 30 4) false)))

  ;; observe that BOMBSPEED is not in scope here, so I had to write (+
  ;; 30 4) instead of (+ 30 BOMBSPEED).  I suppose I could export
  ;; BOMBSPEED via a test method, but that seems inelegant.

  ;; version with internal unit tests
  (send (make-bomb) for-test:run-unit-tests)

)

(run-tests bomb-tests)





