#lang racket

;; 10-6-shapes-with-interfaces.rkt

(require rackunit)
(require rackunit/text-ui)
(require 2htdp/image)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Interface

;; all geometric shapes support these methods in all contexts
;; a Shape is an object of a class that implements Shape<%>.
(define Shape<%> 
  (interface ()

    ;; weight : -> Number
    ;; RETURNS: the weight of this shape
    weight 

    ;; add-to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this shape
    ;; painted on it.
    add-to-scene 

    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Classes

;; A Circle is a (new Circle% [x Number][y Number][r Number][c ColorString])
;; REPRESENTS: a circle 
(define Circle%
  (class* object% (Shape<%>)
    (init-field 
      x                         ; Number, x pixels of center from left
      y                         ; Number, y pixels of center from top
      r                         ; Number, radius
      c)                        ; ColorString 

    (field [IMG (circle r "solid" c)])

    (super-new)

    ;; for each method, we copy down the contract and purpose
    ;; statement from the interface, with perhaps additional details
    ;; relating to this class.

    ;; weight : -> Number
    ;; RETURNS: the weight of this shape
    ;; DETAILS: this shape is a circle
    ;; STRATEGY: domain knowledge
    (define/public (weight) (* pi r r))

    ;; add-to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this shape
    ;; painted on it.
    ;; DETAILS: this shape is a circle
    ;; STRATEGY: domain knowledge
    (define/public (add-to-scene s) (place-image IMG x y s))

    ))

;; A Square is a (new Square% [x Number][y Number][l Number][c ColorString])
;; REPRESENTS: a square parallel to sides of canvas 
(define Square%
  (class* object% (Shape<%>)
    (init-field x  ; Number, x pixels of center from left
                y  ; Number, y pixels of center from top
		l  ; Number, length of one side 
		c) ; ColorString 

    (field [IMG (rectangle l l "solid" c)])

    (super-new)

    ;; weight : -> Number
    ;; RETURNS: the weight of this shape
    ;; DETAILS: this shape is a square
    ;; STRATEGY: domain knowledge
    (define/public (weight) (* l l))

    ;; add-to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this shape
    ;; painted on it.
    ;; DETAILS: this shape is a square
    ;; STRATEGY: domain knowledge
    (define/public (add-to-scene s) (place-image IMG x y s))

    ))

;; A Composite is a (new Composite% [front Shape][back Shape])
;; a composite of front and back
(define Composite%
  (class* object% (Shape<%>)
    (init-field 
      front                             ; Shape, the shape in front
      back                              ; Shape, the shape in back
      )

    (super-new)

    ;; all we know here is that front and back implement Shape<%>.
    ;; we don't know if they are circles, squares, or other composites!

    ;; weight : -> Number
    ;; RETURNS: the weight of this shape
    ;; DETAILS: this shape is a composite
    ;; STRATEGY: structural decomposition on this
    (define/public (weight) (+ (send front weight)
                              (send back weight)))

    ;; add-to-scene : Scene -> Scene
    ;; RETURNS: a scene like the given one, but with this shape
    ;; painted on it.
    ;; DETAILS: this shape is a composite
    ;; strategy: structural decomposition on this
    (define/public (add-to-scene scene)
      (send front add-to-scene
        (send back add-to-scene scene)))
     
    ))

(define EMPTY-CANVAS (empty-scene 200 100))

(define s1 (new Circle% [x 50][y 20][r 40][c "red"]))
(define s2 (new Square% [x 80][y 70][l 40][c "blue"]))
(define s3 (new Composite% [front s1][back s2]))


(define-test-suite shape-tests
  (check-= (send s1 weight) (* pi 1600) .1)
  (check-equal? (send s2 weight) 1600)
  (check-= (send s3 weight) (+ (* pi 1600) 1600) .1))

"shape tests"
(run-tests shape-tests)

